"""
Test Dynamic Raw ID Widgets with Selenium in a Admin Changelist Filterset
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.contrib.auth.models import User
from selenium.webdriver.common.by import By

from dynamic_raw_id.tests.testapp.models import (
    CharPrimaryKeyModel,
    IntPrimaryKeyModel,
    ModelToTest,
    UUIDPrimaryKeyModel,
)

if TYPE_CHECKING:
    from django.db.models import Model
    from selenium.webdriver.firefox.webdriver import WebDriver


@pytest.mark.parametrize(
    ("id_value", "pk_model", "pk_name", "test_values"),
    [
        (
            "dynamic_raw_id_fk",  # Changelist Field ID
            User,  # Primary Key Model to create
            "username",  # Primary Key Model pk fieldname
            ["tick", "trick", "track"],  # Primary Key Test values
        ),
        (
            "dynamic_raw_id_fk_int_pk",
            IntPrimaryKeyModel,
            "num",
            [1, 2, 3],
        ),
        (
            "dynamic_raw_id_fk_char_pk",
            CharPrimaryKeyModel,
            "chr",
            ["abc", "def", "ghi"],
        ),
        (
            "dynamic_raw_id_fk_uuid_pk",
            UUIDPrimaryKeyModel,
            "uuid",
            [
                "95a14df6-8753-41a4-bc7a-94da9f8e82f3",
                "ab34dbb8-d885-48a7-9055-41d02667e4b3",
                "a745c8fc-d39a-4800-89dd-292929379bb3",
            ],
        ),
    ],
)
@pytest.mark.usefixtures("_changelist_page")
def test_changelist_filter(
    selenium: WebDriver,
    id_value: str,
    pk_model: Model,
    pk_name: str,
    test_values: list[str | int],
) -> None:
    """
    Test the Dynamic Raw ID Widget in a Django Admin Changelist Filter.

    This is done for the various types of a primary key (regular, char, uuid, ...).
    """
    # Create testmodel instances
    for value in test_values:
        ModelToTest.objects.create(
            **{id_value: pk_model.objects.create(**{pk_name: value})}
        )

    test_value = test_values[0]

    # Select the "Lookup" icon next to the field
    selenium.find_element(By.ID, f"lookup_id_{id_value}").click()

    #  Activate the popup window with the `window.name = <window_id>`
    selenium.switch_to.window(selenium.window_handles[1])

    # Select the element matching the link text of the first test value
    selenium.find_element(By.LINK_TEXT, str(test_value)).click()

    # Activate the default window
    selenium.switch_to.window(selenium.window_handles[0])

    # Only the first element will be listed on the filtered changelist
    assert selenium.find_element(By.LINK_TEXT, str(test_value)).is_displayed()

    for other_value in test_values[1:]:
        assert len(selenium.find_elements(By.LINK_TEXT, str(other_value))) == 0



"""
Test vanilla Django raw id fields since we overwrite it's Javascript
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from selenium.webdriver.common.by import By

if TYPE_CHECKING:
    from selenium.webdriver.firefox.webdriver import WebDriver


@pytest.mark.usefixtures("_add_page", "_sample_primary_keys")
def test_widgets(selenium: WebDriver) -> None:
    """
    Test the classic Django Foreignkey Widget to make sure,
    this application does not break the vanilla behavior.
    """

    # Input Field ID, Select Value, Test Value
    tests = (
        ("rawid_fk", "user1", "user1"),
        ("rawid_fk_limited", "admin", "admin"),
        ("rawid_many", "user1", None),
        ("rawid_many", "user2", None),  # Second click on Many2Many
    )

    for id_value, select_value, _ in tests:
        # Select the "Lookup" icon next to the field
        selenium.find_element(By.ID, f"lookup_id_{id_value}").click()

        #  Activate the popup window with the `window.name = <window_id>`
        selenium.switch_to.window(selenium.window_handles[1])

        # Select the result element matching the link text
        selenium.find_element(By.LINK_TEXT, select_value).click()

        # Activate the default window
        selenium.switch_to.window(selenium.window_handles[0])

    # Click the save button and wait for page reload
    selenium.find_element(By.NAME, "_continue").click()
    selenium.implicitly_wait(0.5)  # Wait a bit for page reload

    # Test that all selected ForeignKeys are now displayed after reload.
    for id_value, _, test_value in tests:
        # ManyToMany fields don't display a label in Vanilla Django raw id fields.
        if test_value is None:
            continue

        label = selenium.find_element(By.CSS_SELECTOR, f"#id_{id_value} ~ strong")
        assert label.text == test_value



from django.urls import path

from dynamic_raw_id.views import label_view

app_name = "dynamic_raw_id"

urlpatterns = [
    path(
        "<slug:app_name>/<slug:model_name>/multiple/",
        label_view,
        {
            "multi": True,
            "template_object_name": "objects",
            "template_name": "dynamic_raw_id/multi_label.html",
        },
        name="dynamic_raw_id_multi_label",
    ),
    path(
        "<slug:app_name>/<slug:model_name>/",
        label_view,
        {"template_name": "dynamic_raw_id/label.html"},
        name="dynamic_raw_id_label",
    ),
]



"""
Test Dynamic Raw ID Filter with Selenium on a Django Model
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.contrib.auth.models import User
from django.urls import reverse
from selenium.webdriver.common.by import By

from dynamic_raw_id.tests.testapp.models import (
    CharPrimaryKeyModel,
    IntPrimaryKeyModel,
    UUIDPrimaryKeyModel,
)

if TYPE_CHECKING:
    from pytest_django.live_server_helper import LiveServer
    from selenium.webdriver.firefox.webdriver import WebDriver


@pytest.fixture()
def _login(selenium: WebDriver, live_server: LiveServer) -> None:
    """Log into Admin with a Superuser."""
    User.objects.create_superuser("admin", password="admin")  # noqa: S106 Hardcoded password

    selenium.get(f"{live_server}/admin/login/?next=/admin/")
    selenium.find_element(By.CSS_SELECTOR, "input[name=username]").send_keys("admin")
    selenium.find_element(By.CSS_SELECTOR, "input[name=password]").send_keys("admin")
    selenium.find_element(By.CSS_SELECTOR, "input[type=submit]").click()


@pytest.fixture()
def _changelist_page(
    selenium: WebDriver, live_server: LiveServer, _login: None
) -> None:
    """
    Open the "Changelist" admin page of the test model.
    """
    selenium.get(f"{live_server.url}{reverse('admin:testapp_modeltotest_changelist')}")


@pytest.fixture()
def _add_page(selenium: WebDriver, live_server: LiveServer, _login: None) -> None:
    """
    Open the "Add" admin page of the test model.
    """

    selenium.get(f"{live_server.url}{reverse('admin:testapp_modeltotest_add')}")


@pytest.fixture()
def _add_page_with_inlines(
    selenium: WebDriver, live_server: LiveServer, _login: None
) -> None:
    """
    Open the "Add" admin page of the test model.
    """

    selenium.get(
        f"{live_server.url}{reverse('admin:testapp_modeltotestinlinesbase_add')}"
    )


@pytest.fixture()
def _sample_primary_keys() -> None:
    """Create sample Primary key model instances."""
    User.objects.create(username="user1")
    User.objects.create(username="user2")
    IntPrimaryKeyModel.objects.create(num=123)
    CharPrimaryKeyModel.objects.create(chr="abc")
    UUIDPrimaryKeyModel.objects.create(uuid="9a10987b-51ba-472f-9dfe-175286e2258a")



# Generated by Django 5.0.4 on 2024-04-21 15:12

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CharPrimaryKeyModel",
            fields=[
                (
                    "chr",
                    models.CharField(max_length=20, primary_key=True, serialize=False),
                ),
            ],
        ),
        migrations.CreateModel(
            name="IntPrimaryKeyModel",
            fields=[
                (
                    "num",
                    models.IntegerField(
                        primary_key=True, serialize=False, verbose_name="Number"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelToTestInlinesBase",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UUIDPrimaryKeyModel",
            fields=[
                ("uuid", models.UUIDField(primary_key=True, serialize=False)),
            ],
        ),
        migrations.CreateModel(
            name="ModelToTestInlines",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "dynamic_raw_id_fk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="inline_dynamic_raw_id_fk",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Dynamic RawID ForeignKey",
                    ),
                ),
                (
                    "base",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="testapp.modeltotestinlinesbase",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelToTest",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "dynamic_raw_id_fk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dynamic_raw_id_fk",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Dynamic RawID ForeignKey",
                    ),
                ),
                (
                    "dynamic_raw_id_fk_char_pk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dynamic_raw_id_fk_char_pk",
                        to="testapp.charprimarykeymodel",
                        verbose_name="Dynamic RawID Custom Primary Key: Character Field",
                    ),
                ),
                (
                    "dynamic_raw_id_fk_int_pk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dynamic_raw_id_fk_int_pk",
                        to="testapp.intprimarykeymodel",
                        verbose_name="Dynamic RawID Custom Primary Key: Integer Field",
                    ),
                ),
                (
                    "dynamic_raw_id_fk_limited",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to={"is_staff": True},
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dynamic_raw_id_fk_limited",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Dynamic RawID ForeignKey with limited choices",
                    ),
                ),
                (
                    "dynamic_raw_id_many",
                    models.ManyToManyField(
                        blank=True,
                        related_name="dynamic_raw_id_many",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Dynamic RawID ManyToMany",
                    ),
                ),
                (
                    "rawid_fk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rawid_fk",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Regular RawID ForeignKey",
                    ),
                ),
                (
                    "rawid_fk_limited",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to={"is_staff": True},
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rawid_fk_limited",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Regular RawID ForeignKey with limited choices",
                    ),
                ),
                (
                    "rawid_many",
                    models.ManyToManyField(
                        blank=True,
                        related_name="rawid_many",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Regular RawID ManyToMany",
                    ),
                ),
                (
                    "dynamic_raw_id_fk_uuid_pk",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dynamic_raw_id_fk_uuid_pk",
                        to="testapp.uuidprimarykeymodel",
                        verbose_name="Dynamic RawID Custom Primary Key: UUID Field",
                    ),
                ),
            ],
        ),
    ]



"""dynamic_raw_id filters."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django import forms
from django.contrib import admin

from .widgets import DynamicRawIDWidget

if TYPE_CHECKING:
    from django.db.models import Field as DbField
    from django.db.models import Model, QuerySet
    from django.forms import Form
    from django.http import HttpRequest


class DynamicRawIDFilterForm(forms.Form):
    """Form for dynamic_raw_id filter."""

    def __init__(
        self,
        rel: DbField,
        admin_site: admin.AdminSite,
        field_name: str,
        **kwargs: Any,
    ) -> None:
        """Construct field for given field rel."""
        super().__init__(**kwargs)
        self.fields[field_name] = forms.CharField(
            label="",
            widget=DynamicRawIDWidget(rel=rel, admin_site=admin_site),
            required=False,
        )


class DynamicRawIDFilter(admin.filters.FieldListFilter):
    """Filter list queryset by primary key of a related object."""

    template = "dynamic_raw_id/admin/filters/dynamic_raw_id_filter.html"

    def __init__(  # noqa: PLR0913 Too Many arguments
        self,
        field: DbField,
        request: HttpRequest,
        params: dict[str, Any],
        model: Model,
        model_admin: admin.ModelAdmin,
        field_path: str,
    ) -> None:
        """Use GET param for lookup and form initialization."""
        self.lookup_kwarg = field_path
        super().__init__(field, request, params, model, model_admin, field_path)
        rel = field.remote_field
        self.form = self.get_form(request, rel, model_admin.admin_site)

    def choices(self, changelist: Any) -> list:
        """Filter choices do not exist, since we choose the popup value."""
        return []

    def expected_parameters(self) -> str:
        """Return GET params for this filter."""
        return self.lookup_kwarg

    def get_form(
        self,
        request: HttpRequest,
        rel: DbField,
        admin_site: admin.AdminSite,
    ) -> Form:
        """Return a filter form."""
        return DynamicRawIDFilterForm(
            admin_site=admin_site,
            rel=rel,
            field_name=self.field_path,
            data=self.used_parameters,
        )

    def queryset(
        self,
        request: HttpRequest,
        queryset: QuerySet,
    ) -> QuerySet:
        """Filter queryset using params from the form."""
        if self.form.is_valid():
            # get no null params
            filter_params = dict(
                filter(lambda x: bool(x[1]), self.form.cleaned_data.items())
            )
            return queryset.filter(**filter_params)
        return queryset



"""
Test the basic integrity of the app. We can't test the Javascript side
and all those fancy popups, but we can load an admin view and check that
dynamic_raw_id was successfully loaded and displays items properly.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
from django.contrib.auth.models import User
from django.urls import reverse
from pytest_django.asserts import assertContains

from dynamic_raw_id.tests.testapp.models import (
    CharPrimaryKeyModel,
    IntPrimaryKeyModel,
    ModelToTest,
    UUIDPrimaryKeyModel,
)

if TYPE_CHECKING:
    from django.test import Client


@pytest.fixture()
def staff_user(db: Any) -> User:
    """Staff user with no additional permissions."""
    user = User.objects.create_user("staff", "", "staff")
    user.is_staff = True
    user.save()
    return user


@pytest.fixture()
def sample_obj(db: Any) -> dict[str, Any]:
    """
    Create a bit of sample data we can use to assign.
    """
    user1 = User.objects.create_superuser("jon", "jon@example.com", "")
    user2 = User.objects.create_superuser("jim", "jim@example.com", "")
    char = CharPrimaryKeyModel.objects.create(chr="helloworld")
    num = IntPrimaryKeyModel.objects.create(num=12345)
    uuid = UUIDPrimaryKeyModel.objects.create(
        uuid="9a10987b-51ba-472f-9dfe-175286e2258a"
    )

    obj = ModelToTest.objects.create(
        rawid_fk=user1,
        rawid_fk_limited=user1,
        dynamic_raw_id_fk=user1,
        dynamic_raw_id_fk_limited=user1,
        dynamic_raw_id_fk_int_pk=num,
        dynamic_raw_id_fk_char_pk=char,
        dynamic_raw_id_fk_uuid_pk=uuid,
    )
    obj.rawid_many.add(user1, user2)
    obj.dynamic_raw_id_many.add(user1, user2)

    return obj


def get_labelview_url(multi: bool = False) -> str:
    """
    Create a URL to the JSON view that creates the dynamic labels.
    """ ""
    name = multi and "dynamic_raw_id_multi_label" or "dynamic_raw_id_label"
    return reverse(
        f"dynamic_raw_id:{name}",
        kwargs={"app_name": "testapp", "model_name": "modeltotest"},
    )


def test_changelist_integrity(admin_client: Client, sample_obj: Any) -> None:
    """
    The `DynamicRawIDFilter` is hooked up in the right filter bar of the
    testapp changelist view.
    """
    list_url = reverse("admin:testapp_modeltotest_changelist")
    response = admin_client.get(list_url, follow=True)
    assert response.status_code == 200
    assertContains(response, "dynamic_raw_id-related-lookup")


def test_change_integrity(admin_client: Client, sample_obj: Any) -> None:
    """
    The `DynamicRawIDFilter` buttons are displayed next to the FK Widget.
    """
    list_url = reverse("admin:testapp_modeltotest_change", args=(sample_obj.pk,))
    response = admin_client.get(list_url, follow=True)
    assert response.status_code == 200
    assertContains(response, "dynamic_raw_id-clear-field")
    assertContains(response, "dynamic_raw_id-related-lookup")


def test_labelview_unauthed(client: Client, sample_obj: Any) -> None:
    """
    Label view required authentication and a staff account
    """
    response = client.get(get_labelview_url(), follow=True)
    assert response.status_code == 404


def test_labelview_no_permission(
    client: Client, staff_user: User, sample_obj: Any
) -> None:
    """
    Valid Label view request is denied if user has no change permisson for the app.
    """
    client.login(username="staff", password="staff")  # noqa: S106 Hardcoded password
    response = client.get(
        get_labelview_url(multi=True), {"id": sample_obj.pk}, follow=True
    )
    assert response.status_code == 403


def test_labelview(admin_client: Client, sample_obj: Any) -> None:
    """
    Call the labelview directly (what usually an Ajax JS call would do)
    and check for proper response.
    """
    response = admin_client.get(get_labelview_url(), {"id": sample_obj.pk}, follow=True)
    assert response.status_code == 200


def test_multi_labelview(admin_client: Client, sample_obj: Any) -> None:
    """
    Call the labelview directly (what usually an Ajax JS call would do)
    and check for proper response. Exect a multi response.
    """
    response = admin_client.get(
        get_labelview_url(multi=True), {"id": sample_obj.pk}, follow=True
    )
    assert response.status_code == 200


def test_invalid_id(admin_client: Client) -> None:
    """
    Test label view with where model primary key is invalid.
    """
    response = admin_client.get(get_labelview_url(), {"id": "wrong"}, follow=True)
    assert response.status_code == 400


def test_id_does_not_exist(admin_client: Client) -> None:
    """
    Test label view with where a model primary key does not exist.
    """
    response = admin_client.get(get_labelview_url(), {"id": "123456"}, follow=True)
    assert response.status_code == 400


def test_no_id(admin_client: Client) -> None:
    """
    Test label view with no ID given.
    """
    response = admin_client.get(get_labelview_url(), follow=True)
    assert response.status_code == 400


def test_invalid_appname(admin_client: Client) -> None:
    """
    Test label view with invalid app/model name.
    """
    url = get_labelview_url().replace("testapp", "foobar")
    response = admin_client.get(url, {"id": "123456"}, follow=True)
    assert response.status_code == 400



"""
Test Dynamic Raw ID Widgets with Selenium in a ModelAdmin
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.contrib.auth.models import User
from selenium.webdriver.common.by import By

from dynamic_raw_id.tests.testapp.models import (
    CharPrimaryKeyModel,
    IntPrimaryKeyModel,
    UUIDPrimaryKeyModel,
)

if TYPE_CHECKING:
    from selenium.webdriver.firefox.webdriver import WebDriver


@pytest.mark.usefixtures("_add_page", "_sample_primary_keys")
def test_widgets(selenium: WebDriver) -> None:
    """
    Test the variations of the Dynamic Raw ID Widget.

    The tests select the "Lookup" icon, choose the first element in the result set,
    and test that the selected element name appears next to the field.

    Doing this in a all-in-one testcase, rather a parameterized test, so it's much
    faster.
    """

    # Input Field ID, Select Value, Test Value
    tests = (
        ("dynamic_raw_id_fk", "user1", "user1"),
        ("dynamic_raw_id_fk_limited", "admin", "admin"),
        ("dynamic_raw_id_many", "user1", "user1"),
        ("dynamic_raw_id_many", "user2", "user1,  user2"),  # Second click on Many2Many
        ("dynamic_raw_id_fk_int_pk", "123", "123"),
        ("dynamic_raw_id_fk_char_pk", "abc", "abc"),
        (
            "dynamic_raw_id_fk_uuid_pk",
            "9a10987b-51ba-472f-9dfe-175286e2258a",
            "9a10987b-51ba-472f-9dfe-175286e2258a",
        ),
    )

    for id_value, select_value, test_value in tests:
        # Select the "Lookup" icon next to the field
        selenium.find_element(By.ID, f"lookup_id_{id_value}").click()

        #  Activate the popup window with the `window.name = <window_id>`
        selenium.switch_to.window(selenium.window_handles[1])

        # Select the result element matching the link text
        selenium.find_element(By.LINK_TEXT, select_value).click()

        # Activate the default window
        selenium.switch_to.window(selenium.window_handles[0])

        # Find the attached label next to it
        label = selenium.find_element(By.ID, f"{id_value}_dynamic_raw_id_label")

        # Test that the displayed value matches the test value
        selenium.implicitly_wait(0.2)  # Wait a bit for page reload
        assert label.text == test_value

        # For debugging, draw green border around the label and wait a bit
        # selenium.execute_script("arguments[0].style.border='3px solid green'", label)

    # Save the change form once and make sure, the previously
    # entered values are shown correctly.
    selenium.find_element(By.NAME, "_continue").click()
    selenium.implicitly_wait(0.5)  # Wait a bit for page reload

    # Check values are all intact
    el = selenium.find_element(By.ID, "id_dynamic_raw_id_fk")
    assert el.get_attribute("value") == str(User.objects.get(username="user1").pk)

    el = selenium.find_element(By.ID, "id_dynamic_raw_id_fk_limited")
    assert el.get_attribute("value") == str(User.objects.get(username="admin").pk)

    el = selenium.find_element(By.ID, "id_dynamic_raw_id_many")
    u1 = User.objects.get(username="user1").pk
    u2 = User.objects.get(username="user2").pk
    assert el.get_attribute("value") == f"{u1},{u2}"

    el = selenium.find_element(By.ID, "id_dynamic_raw_id_fk_int_pk")
    assert el.get_attribute("value") == str(IntPrimaryKeyModel.objects.get(num=123).pk)

    el = selenium.find_element(By.ID, "id_dynamic_raw_id_fk_char_pk")
    assert el.get_attribute("value") == str(
        CharPrimaryKeyModel.objects.get(chr="abc").pk
    )

    el = selenium.find_element(By.ID, "id_dynamic_raw_id_fk_uuid_pk")
    assert el.get_attribute("value") == str(
        UUIDPrimaryKeyModel.objects.get(uuid="9a10987b-51ba-472f-9dfe-175286e2258a").pk
    )



from django.apps import apps
from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
)
from django.shortcuts import render
from django.urls import reverse


@user_passes_test(lambda u: u.is_staff)
def label_view(  # noqa: PLR0913 Too Many arguments
    request: HttpRequest,
    app_name: str,
    model_name: str,
    template_name: str,
    multi: bool = False,
    template_object_name: str = "object",
) -> HttpResponse:
    # The list of to obtained objects is in GET.id. No need to resume if we
    # didn't get it.
    if not request.GET.get("id"):
        msg = "No list of objects given"
        return HttpResponseBadRequest(settings.DEBUG and msg or "")

    # Given objects are either an integer or a comma-separated list of
    # integers. Validate them and ignore invalid values. Also strip them
    # in case the user entered values by hand, such as '1, 2,3'.
    object_list = [pk.strip() for pk in request.GET["id"].split(",")]

    # Make sure this model exists and the user has 'change' permission for it.
    # If he doesn't have this permission, Django would not display the
    # change_list in the popup, and the user was never able to select objects.
    try:
        model = apps.get_model(app_name, model_name)
    except LookupError:
        msg = f"Model {app_name}.{model_name} does not exist."
        return HttpResponseBadRequest(msg)

    # Check 'view' permission
    if not request.user.has_perm(f"{app_name}.view_{model_name}"):
        return HttpResponseForbidden()

    try:
        if multi:
            model_template = f"dynamic_raw_id/{app_name}/multi_{model_name}.html"
            objs = model.objects.filter(pk__in=object_list)
            objects = [
                (obj, reverse(f"admin:{app_name}_{model_name}_change", args=[obj.pk]))
                for obj in objs
            ]
            extra_context = {template_object_name: objects}
        else:
            model_template = f"dynamic_raw_id/{app_name}/{model_name}.html"
            obj = model.objects.get(pk=object_list[0])
            change_url = reverse(f"admin:{app_name}_{model_name}_change", args=[obj.pk])
            extra_context = {template_object_name: (obj, change_url)}

    # most likely, the pk wasn't convertable
    except ValueError:
        msg = "ValueError during lookup"
        return HttpResponseBadRequest(msg)
    except model.DoesNotExist:
        msg = "Model instance does not exist"
        return HttpResponseBadRequest(msg)

    return render(request, (model_template, template_name), extra_context)



from tempfile import TemporaryDirectory

DEBUG = True

SECRET_KEY = "super-secret-dynamic_raw_id-key"  # noqa: S105 Hardcoded password


ROOT_URLCONF = "dynamic_raw_id.tests.testapp.urls"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "testapp.db"},
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.messages",
    "dynamic_raw_id",
    "dynamic_raw_id.tests.testapp",
]

MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)


STATIC_ROOT = TemporaryDirectory().name
STATIC_URL = "/static/"

STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]

USE_TZ = False



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.contrib.admin.options import BaseModelAdmin

from .widgets import DynamicRawIDMultiIdWidget, DynamicRawIDWidget

if TYPE_CHECKING:
    from django.db.models.fields import Field as DB_Field
    from django.forms.fields import Field as FormField
    from django.http import HttpRequest


class DynamicRawIDMixin(BaseModelAdmin):
    dynamic_raw_id_fields = ()

    def formfield_for_foreignkey(
        self,
        db_field: DB_Field,
        request: HttpRequest | None = None,
        **kwargs: Any,
    ) -> FormField:
        if db_field.name in self.dynamic_raw_id_fields:
            rel = db_field.remote_field
            kwargs["widget"] = DynamicRawIDWidget(rel, self.admin_site)
            kwargs["help_text"] = ""
            return db_field.formfield(**kwargs)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(
        self,
        db_field: DB_Field,
        request: HttpRequest | None = None,
        **kwargs: Any,
    ) -> FormField:
        if db_field.name in self.dynamic_raw_id_fields:
            rel = db_field.remote_field
            kwargs["widget"] = DynamicRawIDMultiIdWidget(rel, self.admin_site)
            kwargs["help_text"] = ""
            return db_field.formfield(**kwargs)
        return super().formfield_for_manytomany(db_field, request, **kwargs)



from django.db import models


# Test custom forms of primary key fields
class IntPrimaryKeyModel(models.Model):
    num = models.IntegerField("Number", primary_key=True)

    def __str__(self) -> str:
        return str(self.num)


class CharPrimaryKeyModel(models.Model):
    chr = models.CharField(max_length=20, primary_key=True)

    def __str__(self) -> str:
        return str(self.chr)


class UUIDPrimaryKeyModel(models.Model):
    uuid = models.UUIDField(primary_key=True)

    def __str__(self) -> str:
        return str(self.uuid)


class ModelToTest(models.Model):
    # Regular RawID Fields -------------------------------------------------------------
    rawid_fk = models.ForeignKey(
        "auth.User",
        related_name="rawid_fk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Regular RawID ForeignKey",
    )
    rawid_fk_limited = models.ForeignKey(
        "auth.User",
        related_name="rawid_fk_limited",
        limit_choices_to={"is_staff": True},
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Regular RawID ForeignKey with limited choices",
    )
    rawid_many = models.ManyToManyField(
        "auth.User",
        related_name="rawid_many",
        blank=True,
        verbose_name="Regular RawID ManyToMany",
    )

    # Dynamic RawID Fields with Django Default PK --------------------------------------
    dynamic_raw_id_fk = models.ForeignKey(
        "auth.User",
        related_name="dynamic_raw_id_fk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID ForeignKey",
    )
    dynamic_raw_id_fk_limited = models.ForeignKey(
        "auth.User",
        related_name="dynamic_raw_id_fk_limited",
        limit_choices_to={"is_staff": True},
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID ForeignKey with limited choices",
    )
    dynamic_raw_id_many = models.ManyToManyField(
        "auth.User",
        related_name="dynamic_raw_id_many",
        blank=True,
        verbose_name="Dynamic RawID ManyToMany",
    )

    # Dynamic RawID Fields with Custom PK Fields ---------------------------------------
    # Custom PrimaryKey Field Model with a Integer value
    dynamic_raw_id_fk_int_pk = models.ForeignKey(
        IntPrimaryKeyModel,
        related_name="dynamic_raw_id_fk_int_pk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID Custom Primary Key: Integer Field",
    )

    # Custom PrimaryKey Field Model with a Character value
    dynamic_raw_id_fk_char_pk = models.ForeignKey(
        CharPrimaryKeyModel,
        related_name="dynamic_raw_id_fk_char_pk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID Custom Primary Key: Character Field",
    )

    # Custom PrimaryKey Field Model with an UUID value
    dynamic_raw_id_fk_uuid_pk = models.ForeignKey(
        UUIDPrimaryKeyModel,
        related_name="dynamic_raw_id_fk_uuid_pk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID Custom Primary Key: UUID Field",
    )

    def __str__(self) -> str:
        if self.dynamic_raw_id_fk:
            return str(self.dynamic_raw_id_fk)
        if self.dynamic_raw_id_fk_int_pk:
            return str(self.dynamic_raw_id_fk_int_pk)
        if self.dynamic_raw_id_fk_char_pk:
            return str(self.dynamic_raw_id_fk_char_pk)
        if self.dynamic_raw_id_fk_uuid_pk:
            return str(self.dynamic_raw_id_fk_uuid_pk)
        return "Test Instance"


# Test Inline Admin Model --------------------------------------------------------------
class ModelToTestInlinesBase(models.Model):
    def __str__(self) -> str:
        return "Base Model with Inlines"


class ModelToTestInlines(models.Model):
    base = models.ForeignKey(ModelToTestInlinesBase, on_delete=models.CASCADE)

    dynamic_raw_id_fk = models.ForeignKey(
        "auth.User",
        related_name="inline_dynamic_raw_id_fk",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name="Dynamic RawID ForeignKey",
    )

    def __str__(self) -> str:
        return "Inline Model"



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urlencode

from django import forms
from django.contrib.admin import widgets
from django.urls import reverse
from django.utils.encoding import force_str

if TYPE_CHECKING:
    from django.forms.renderers import BaseRenderer
    from django.template import Context


class DynamicRawIDWidget(widgets.ForeignKeyRawIdWidget):
    template_name: str = "dynamic_raw_id/admin/widgets/dynamic_raw_id_field.html"

    def get_context(self, name: str, value: Any, attrs: dict[str, Any]) -> Context:
        context = super().get_context(name, value, attrs)
        app_name = self.rel.model._meta.app_label  # noqa: SLF001 Private member accessed
        model_name = self.rel.model._meta.object_name.lower()  # noqa: SLF001 Private member accessed

        attrs.setdefault("class", "vForeignKeyRawIdAdminField")

        context.update(
            name=name,
            app_name=app_name,
            model_name=model_name,
            related_url=reverse(
                f"admin:{app_name}_{model_name}_changelist",
                current_app=self.admin_site.name,
            ),
            url=f"?{urlencode(self.url_parameters())}",
        )
        return context

    @property
    def media(self) -> forms.Media:
        return forms.Media(
            js=[
                "admin/js/vendor/jquery/jquery.min.js",
                "admin/js/jquery.init.js",
                "admin/js/core.js",
                "dynamic_raw_id/js/dynamic_raw_id.js",
            ]
        )


class DynamicRawIDMultiIdWidget(DynamicRawIDWidget):
    def value_from_datadict(
        self,
        data: dict[str, Any],
        files: Any | None,
        name: str,
    ) -> str | None:
        value = data.get(name)
        if value:
            return value.split(",")
        return None

    def render(
        self,
        name: str,
        value: Any,
        attrs: dict[str, Any] | None = None,
        renderer: BaseRenderer | None = None,
    ) -> str:
        attrs["class"] = "vManyToManyRawIdAdminField"
        value = ",".join([force_str(v) for v in value]) if value else ""
        return super().render(name, value, attrs, renderer=renderer)



from django.contrib import admin

from dynamic_raw_id.admin import DynamicRawIDMixin
from dynamic_raw_id.filters import DynamicRawIDFilter

from . import models


@admin.register(models.ModelToTest)
class ModelToTestAdmin(DynamicRawIDMixin, admin.ModelAdmin):
    raw_id_fields = (
        "rawid_fk",
        "rawid_fk_limited",
        "rawid_many",
    )
    dynamic_raw_id_fields = (
        "dynamic_raw_id_fk",
        "dynamic_raw_id_fk_limited",
        "dynamic_raw_id_many",
        "dynamic_raw_id_fk_int_pk",
        "dynamic_raw_id_fk_char_pk",
        "dynamic_raw_id_fk_uuid_pk",
    )
    list_filter = (
        ("dynamic_raw_id_fk", DynamicRawIDFilter),
        ("dynamic_raw_id_fk_int_pk", DynamicRawIDFilter),
        ("dynamic_raw_id_fk_char_pk", DynamicRawIDFilter),
        ("dynamic_raw_id_fk_uuid_pk", DynamicRawIDFilter),
    )


class ModelToTestInlineAdmin(DynamicRawIDMixin, admin.TabularInline):
    model = models.ModelToTestInlines
    dynamic_raw_id_fields = (
        "dynamic_raw_id_fk",
        "dynamic_raw_id_fk_limited",
        "dynamic_raw_id_many",
        "dynamic_raw_id_fk_int_pk",
        "dynamic_raw_id_fk_char_pk",
        "dynamic_raw_id_fk_uuid_pk",
    )


@admin.register(models.ModelToTestInlinesBase)
class ModelToTestInlineBaseAdmin(admin.ModelAdmin):
    inlines = (ModelToTestInlineAdmin,)


admin.site.register(models.IntPrimaryKeyModel)
admin.site.register(models.CharPrimaryKeyModel)
admin.site.register(models.UUIDPrimaryKeyModel)



"""
Test Dynamic Raw ID Widgets with Selenium in a ModelAdmin Inline
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from django.contrib.auth.models import User
from selenium.webdriver.common.by import By

if TYPE_CHECKING:
    from selenium.webdriver.firefox.webdriver import WebDriver


@pytest.mark.usefixtures("_add_page_with_inlines", "_sample_primary_keys")
def test_widgets(selenium: WebDriver) -> None:
    """
    Django Admin adds three 'inline' rows to the base model which we fill with
    three users.

    Then click the 'Add another' link three times, and fill another three users.
    """

    # Create six test users first
    for uid in range(6):
        test_value = f"inline{uid}"
        User.objects.create_user(username=test_value)

        # The inline form field starts with a zero index and is named
        # `lookup_id_modeltotestinlines_set-0-dynamic_raw_id_fk`

        # If this is the fourth user, click the 'Add another' link three times.
        if uid == 3:
            for _ in range(3):
                selenium.find_element(
                    By.LINK_TEXT, "Add another Model to test inlines"
                ).click()

        # Select the "Lookup" icon next to the field
        selenium.find_element(
            By.ID, f"lookup_id_modeltotestinlines_set-{uid}-dynamic_raw_id_fk"
        ).click()

        #  Activate the popup window with the `window.name = <window_id>`
        selenium.switch_to.window(selenium.window_handles[1])

        # Select the result element matching the link text
        selenium.find_element(By.LINK_TEXT, test_value).click()

        # Activate the default window
        selenium.switch_to.window(selenium.window_handles[0])

        # Find the attached label next to it, and draw a red border around it
        label = selenium.find_element(
            By.ID,
            f"modeltotestinlines_set-{uid}-dynamic_raw_id_fk_dynamic_raw_id_label",
        )

        # Test that the displayed value matches the test value
        assert label.text == test_value
